/// @docImport 'package:consoledoctor_barcode_scanner/src/barcode_scanner_controller.dart';
library;

import 'dart:ui';

import 'package:consoledoctor_barcode_scanner/src/enums/camera_facing.dart';
import 'package:consoledoctor_barcode_scanner/src/enums/barcode_scanner_error_code.dart';
import 'package:consoledoctor_barcode_scanner/src/enums/torch_state.dart';
import 'package:consoledoctor_barcode_scanner/src/barcode_scanner_exception.dart';

/// This class represents the current state of a [BarcodeScannerController].
class BarcodeScannerState {
  /// Create a new [BarcodeScannerState] instance.
  const BarcodeScannerState({
    required this.availableCameras,
    required this.cameraDirection,
    required this.isInitialized,
    required this.isRunning,
    required this.size,
    required this.torchState,
    required this.zoomScale,
    required this.exposureLevel,
    this.minExposureLevel = 0.0,
    this.maxExposureLevel = 0.0,
    this.maxZoomLevel = 1.0,
    this.error,
  });

  /// Create a new [BarcodeScannerState] instance that is uninitialized.
  const BarcodeScannerState.uninitialized(CameraFacing facing)
      : this(
          availableCameras: null,
          cameraDirection: facing,
          isInitialized: false,
          isRunning: false,
          size: Size.zero,
          torchState: TorchState.unavailable,
          zoomScale: 1.0,
          exposureLevel: 0.0,
          minExposureLevel: 0.0,
          maxExposureLevel: 0.0,
          maxZoomLevel: 1.0
        );

  /// The number of available cameras.
  ///
  /// This is null if the number of cameras is unknown.
  final int? availableCameras;

  /// The facing direction of the camera.
  final CameraFacing cameraDirection;

  /// The error that occurred while setting up or using the camera.
  final BarcodeScannerException? error;

  /// Whether the mobile scanner has initialized successfully.
  ///
  /// This does not indicate that the camera permission was granted.
  /// To check if the camera permission was granted, use [hasCameraPermission].
  final bool isInitialized;

  /// Whether the mobile scanner is currently running.
  ///
  /// This is `true` if the camera is active.
  final bool isRunning;

  /// The size of the camera output.
  final Size size;

  /// The current state of the flashlight of the camera.
  final TorchState torchState;

  /// The current zoom scale of the camera.
  final double zoomScale;

  /// The minimum exposure level of the camera.
  final double minExposureLevel;

  /// The maximum exposure level of the camera.
  final double maxExposureLevel;

  /// The current exposure level of the camera.
  final double exposureLevel;

  /// The max zoom level of the camera.
  final double maxZoomLevel;

  /// Whether permission to access the camera was granted.
  bool get hasCameraPermission {
    return isInitialized &&
        error?.errorCode != BarcodeScannerErrorCode.permissionDenied;
  }

  /// Create a copy of this state with the given parameters.
  BarcodeScannerState copyWith({
    int? availableCameras,
    CameraFacing? cameraDirection,
    BarcodeScannerException? error,
    bool? isInitialized,
    bool? isRunning,
    Size? size,
    TorchState? torchState,
    double? zoomScale,
    double? exposureLevel,
    double? minExposureLevel,
    double? maxExposureLevel,
    double? maxZoomLevel,
    bool resetError = false
  }) {
    return BarcodeScannerState(
      availableCameras: availableCameras ?? this.availableCameras,
      cameraDirection: cameraDirection ?? this.cameraDirection,
      error: resetError ? null : error ?? this.error,
      isInitialized: isInitialized ?? this.isInitialized,
      isRunning: isRunning ?? this.isRunning,
      size: size ?? this.size,
      torchState: torchState ?? this.torchState,
      zoomScale: zoomScale ?? this.zoomScale,
      exposureLevel: exposureLevel ?? this.exposureLevel,
      minExposureLevel: minExposureLevel ?? this.minExposureLevel,
      maxExposureLevel: maxExposureLevel ?? this.maxExposureLevel,
      maxZoomLevel: maxZoomLevel ?? this.maxZoomLevel,
    );
  }
}
