import 'dart:async';
import 'dart:math';

import 'package:consoledoctor_barcode_scanner/consoledoctor_barcode_scanner.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';


// ignore: must_be_immutable
class BarcodeScannerWidget extends StatefulWidget {
  BarcodeScannerWidget({
    super.key,
    this.overlay,
    this.scanDelay = 250,
    this.detectInvertedCodes = false,
    this.initialScanningMode = ScanningMode.pressAndHold,
    List<Widget>? extraWidgets
  }) {
    this.extraWidgets = extraWidgets ?? [];
  }

  final ScanningMode initialScanningMode;
  final ScanOverlayWidget? overlay;
  final int scanDelay;
  final bool detectInvertedCodes;
  late final List<Widget>? extraWidgets;

  State? _currentState;
  void triggerScanEvent(BarcodeCapture capture) {
    if (_currentState != null) {
      (_currentState as _BarcodeScannerWidgetState).onBarcodeDetected(capture);
    }
  }

  @override
  State<BarcodeScannerWidget> createState() =>
      _BarcodeScannerWidgetState();
}

class _BarcodeScannerWidgetState extends State<BarcodeScannerWidget> with WidgetsBindingObserver {

  final double _cameraBtnSize = 74;
  final double _cameraFlipBtnSize = 56;
  double _cameraBtnActiveSize() => _cameraBtnSize * 0.66;

  bool _exposureControlVisible = false;

  bool _showOverlay = false;

  bool _cameraButtonPressed = false;
  int _scanningMode = ScanningMode.pressAndHold.index;
  final List<bool> _selectedScanningModes = <bool>[false, false, false];

  late final BarcodeScannerController controller;

  StreamSubscription<BarcodeCapture?>? _barcodesSubscription;

  @override
  void initState() {
    super.initState();
    widget._currentState = this;

    _initializeScanningMode();
    controller = BarcodeScannerController(
      autoStart: false,
      flashEnabled: false,
      detectionTimeoutMs: widget.scanDelay,
      detectionSpeed: _showOverlay || !_cameraButtonPressed
          ? DetectionSpeed.disabled : DetectionSpeed.normal,
      tryInverted: widget.detectInvertedCodes
    );

    WidgetsBinding.instance.addObserver(this);
    unawaited(_setupBarcodeSubscriptionAndStart());
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (!controller.value.hasCameraPermission) {
      return;
    }

    switch (state) {
      case AppLifecycleState.detached:
      case AppLifecycleState.hidden:
      case AppLifecycleState.paused:
        return;
      case AppLifecycleState.resumed:
        unawaited(controller.start());
      case AppLifecycleState.inactive:
        unawaited(controller.stop());
    }
  }

  Future<void> _setupBarcodeSubscriptionAndStart() {
    _barcodesSubscription = BarcodeScannerPlatform.instance.barcodesStream.listen(
          (BarcodeCapture? barcode) {
        if (barcode == null) {
          return;
        }
        if (!_showOverlay) {
          onBarcodeDetected(barcode);
        }
      },
      onError: (Object error) {
        return;
      },
      cancelOnError: false,
    );

    _initializeOverlay();

    return controller.start();
  }

  void _initializeOverlay() {
    widget.overlay?.dispatcher.onCloseRequested(() {
      setState(() {
        _showOverlay = false;
      });
      unawaited(updateDetectionSpeed());
    });
  }

  void _initializeScanningMode() {
    var scanModeIndex = widget.initialScanningMode.index;
    _cameraButtonPressed = widget.initialScanningMode == ScanningMode.continuous;
    _scanningMode = scanModeIndex;
    _selectedScanningModes[scanModeIndex] = true;
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      backgroundColor: Colors.grey.shade600,
      body: Stack(
        fit: StackFit.expand,
        children: [
          _getFeed(),
          if (kDebugMode)
            BarcodeOverlay(controller: controller, boxFit: BoxFit.contain),
          CameraPinchZoomDetector(controller: controller),
          _backButton(),
          _modeSelector(),
          _flashButton(),
          if (widget.extraWidgets?.isNotEmpty ?? false)
            ...widget.extraWidgets!,
          if (widget.overlay != null)
            _scanResultOverlay()
        ],
      ),
    );
  }

  Widget _scanResultOverlay() => widget.overlay != null
      ? Positioned.fill(child: IgnorePointer(ignoring: !_showOverlay, child: Opacity(opacity: _showOverlay ? 1 : 0, child: widget.overlay)))
      : SizedBox.shrink();

  Widget _getFeed() {
    return BarcodeScanner(
      controller: controller,
      errorBuilder: (context, error, child) {
        return ScannerErrorWidget(error: error);
      },
    );
  }

  Widget _backButton() => Navigator.of(context).canPop()
    ? Positioned(
      top: MediaQuery.of(context).padding.top,
      left: 8,
      child: SizedBox(
        height: 48,
        width: 48,
        child:IconButton(
          icon: Icon(Icons.arrow_back_ios_outlined, color: Colors.white, size: 26),
          onPressed: () => Navigator.of(context).pop(),
        ),
      ),
    )
    : SizedBox();

  Widget _flashButton() => Positioned(
    top: MediaQuery.of(context).padding.top,
    right: 8,
    child: ToggleFlashlightButton(controller: controller),
  );

  Widget _modeSelector() => Positioned(
      bottom: 0,
      right: 0,
      left: 0,
      child: Align(
        alignment: Alignment.bottomCenter,
        child: ColoredBox(
          color: Colors.black.withAlpha(102),
          child: Padding(
            padding: EdgeInsets.fromLTRB(8,12,8, max(MediaQuery.of(context).padding.bottom, 8)),
            child: Column(
                mainAxisAlignment: MainAxisAlignment.center,
                children: [
                  _exposureSlider(),
                  Container(
                    constraints: const BoxConstraints.expand(width: double.infinity, height: 22),
                    child: ToggleButtons(
                      direction: Axis.horizontal,
                      onPressed: (int index) async {
                        setState(() {
                          for (int i = 0; i < _selectedScanningModes.length; i++) {
                            _selectedScanningModes[i] = i == index;
                          }
                          _scanningMode = index;
                          _cameraButtonPressed = index == 0;
                        });

                        await updateDetectionSpeed();
                      },
                      selectedBorderColor: Colors.transparent,
                      selectedColor: Colors.white,
                      fillColor: Colors.transparent,
                      borderColor: Colors.transparent,
                      splashColor: Colors.transparent,
                      renderBorder: false,
                      color: Colors.white30,
                      constraints: BoxConstraints.expand(height: 22, width: (MediaQuery.of(context).size.width - 16) /3,),
                      isSelected: _selectedScanningModes,
                      children: const [
                        Text("CONTINUOUS"),
                        Text("PRESS & HOLD"),
                        Text("TOGGLE")
                      ],
                    ),
                  ),
                  SizedBox.fromSize(size: const Size(1, 12)),
                  Row(
                    children: [
                      Container(width: 8),

                      ValueListenableBuilder(
                          valueListenable: controller,
                          builder: (context, state, child) {
                            if (!state.isInitialized || !state.isRunning) {
                              return const SizedBox.square(dimension: 56);
                            }

                            return RawMaterialButton(
                              onPressed: !state.isInitialized || !state.isRunning || state.minExposureLevel == state.maxExposureLevel ? null : () async {
                                await controller.resetExposureLevel();
                                setState(() {
                                  _exposureControlVisible = !_exposureControlVisible;
                                });
                              },
                              constraints: BoxConstraints(minWidth: 56, maxWidth: 56, minHeight: 56, maxHeight: 56),
                              elevation: 2.0,
                              fillColor: _exposureControlVisible ? Color.fromRGBO(231,76,60,1) : Color.fromRGBO(44, 44, 44, 1),
                              padding: const EdgeInsets.all(0.0),
                              shape: const CircleBorder(),
                              child: const Icon(Icons.exposure, color: Colors.white),
                            );
                          }
                      ),

                      Expanded(child: Container()),

                      GestureDetector(
                        onTapDown: (tap) async{
                          if (_scanningMode != 0) {
                            setState(() {
                              _cameraButtonPressed = !_cameraButtonPressed;
                            });
                            await updateDetectionSpeed();
                          }
                        },
                        onTapUp: (tap) async {
                          if (_scanningMode == 1) {
                            setState(() {
                              _cameraButtonPressed = !_cameraButtonPressed;
                            });
                            await updateDetectionSpeed();
                          }
                        },
                        onTapCancel: () async {
                          if (_scanningMode == 1) {
                            setState(() {
                              _cameraButtonPressed = !_cameraButtonPressed;
                            });
                            await updateDetectionSpeed();
                          }
                        },
                        child: Stack(
                            alignment: Alignment.center,
                            children: [
                              Container(
                                height: 90,
                                width: 90,
                                decoration: BoxDecoration(
                                    border: Border.all(color: Colors.white, width: 4),
                                    color: Colors.transparent,
                                    shape: BoxShape.circle
                                ),
                              ),
                              Visibility(
                                visible: !_cameraButtonPressed,
                                replacement: Container(
                                    constraints: BoxConstraints(minWidth: _cameraBtnActiveSize(), maxWidth: _cameraBtnActiveSize(), minHeight: _cameraBtnActiveSize(), maxHeight: _cameraBtnActiveSize()),
                                    decoration: BoxDecoration(
                                      border: Border.all(color:  const Color.fromRGBO(231, 76, 70, 1), width: 0),
                                      shape: BoxShape.rectangle,
                                      borderRadius: BorderRadius.all(Radius.circular(_cameraBtnActiveSize() /4)),
                                      color: const Color.fromRGBO(231, 76, 70, 1),
                                    )
                                ),
                                child: Container(
                                  height: _cameraBtnSize,
                                  width: _cameraBtnSize,
                                  decoration: BoxDecoration(
                                      border: Border.all(color: Colors.white, width: 0),
                                      shape: BoxShape.circle,
                                      color: Colors.white
                                  ),
                                ),
                              ),
                            ]
                        ),
                      ),
                      Expanded(child: Container()),

                      RawMaterialButton(
                        onPressed: () async {
                          await controller.switchCamera();
                        },
                        constraints: BoxConstraints(minWidth: _cameraFlipBtnSize, maxWidth: _cameraFlipBtnSize, minHeight: _cameraFlipBtnSize, maxHeight: _cameraFlipBtnSize),
                        elevation: 2.0,
                        fillColor: const Color.fromRGBO(44, 44, 44, 1),
                        padding: const EdgeInsets.all(0.0),
                        shape: const CircleBorder(),
                        child: const Icon(Icons.flip_camera_android, color: Colors.white),
                      ),
                      SizedBox(width: 16)
                    ],
                  )
                ]
            ),
          ),
        ),
      )
  );

  Widget _exposureSlider() {
    return !_exposureControlVisible
        ? SizedBox.shrink()
        : ExposureSlider(controller: controller);
  }

  Future<void> updateDetectionSpeed() async {
    await controller.setDetectionSpeed(
        _showOverlay || !_cameraButtonPressed
          ? DetectionSpeed.disabled : DetectionSpeed.normal
    );
  }

  void onBarcodeDetected(BarcodeCapture barcodes) async {
    if (widget.overlay == null) return;

    setState(() {
      _showOverlay = true;
    });
    await updateDetectionSpeed();
    await widget.overlay?.dispatcher.barcodeDetected(barcodes);
  }

  @override
  Future<void> dispose() async {
    WidgetsBinding.instance.removeObserver(this);
    super.dispose();
    await controller.dispose();
    _barcodesSubscription?.cancel();
    _barcodesSubscription = null;
    if (widget._currentState == this) widget._currentState = null;
  }
}
