import 'dart:async';

import 'package:consoledoctor_barcode_scanner/consoledoctor_barcode_scanner.dart';
import 'package:barcode_widget/barcode_widget.dart';
import 'package:consoledoctor_common/consoledoctor_common.dart';
import 'package:flutter/material.dart';
import 'package:flutter_svg/svg.dart';

// ignore: library_prefixes
import 'package:consoledoctor_barcode_scanner/src/objects/barcode.dart' as BarcodeScanner;
// ignore: library_prefixes
import 'package:barcode/barcode.dart' as BarcodeLib;


class ScanResultOverlay extends ScanOverlayWidget {
  ScanResultOverlay({super.key, required super.handler});

  @override
  State createState() => _ScanResultOverlayState();
}

class _ScanResultOverlayState extends State<ScanResultOverlay> {

  String _barcodeText = '0690420';
  BarcodeLib.BarcodeType _barcodeType = BarcodeLib.BarcodeType.Code128;

  ScanOutcomeType _outcomeType = ScanOutcomeType.pending;
  bool _dataAvailable = false;

  final String _defaultLoadingStatusMessage = 'Looking up barcode...';
  String _loadingStatusMessage = 'Looking up barcode...';

  String _heading = '';
  String _subHeading = '';
  String _message = '';
  String? _svg;
  String? _svgPackage;

  @override
  void initState() {
    super.initState();

    _initialize();
  }

  void _initialize() {
    widget.handler.updateStatusMessageCallback = _updateLoadingStatus;
    widget.dispatcher.onBarcodeDetected(processBarcode);
  }

  @override
  void dispose() {
    super.dispose();
  }


  Future<void> processBarcode(BarcodeCapture barcodes) async {
    if (barcodes.barcodes.isEmpty) return;

    BarcodeScanner.Barcode? barcodeScan;
    for (var scan in barcodes.barcodes) {
      if (scan.rawValue?.isNotEmpty ?? false) barcodeScan = scan;
    }

    if (barcodeScan == null) return;

    setState(() {
      _loadingStatusMessage = _defaultLoadingStatusMessage;
      _dataAvailable = false;
      _outcomeType = ScanOutcomeType.pending;
      _barcodeText = barcodeScan!.rawValue!;
      _barcodeType = barcodeScan.format.asBarcodeType();
    });
    var result = await widget.handler.lookupBarcode(barcodes);
    setState(() {
      _dataAvailable = true;
      _outcomeType = result.outcome;
      _heading = result.heading;
      _subHeading = result.subHeading.emptyIfNull();
      _message = result.message;
      _svg = result.svgIcon;
      _svgPackage = result.svgPackage;
    });
  }

  Color _getOverlayColor() => switch (_outcomeType) {
    ScanOutcomeType.valid => Colors.green.shade700,
    ScanOutcomeType.error => BrandColors.dark.accent,
    ScanOutcomeType.warning => ColorExtensions.fromHex('#fb8b01'),
    _ => Colors.black.withValues(alpha: 0.8)
  };

  void _updateLoadingStatus(String message) {
    setState(() {
      _loadingStatusMessage = message;
    });
  }

  List<Widget> _renderOutcomeWidgets() {
    return [
      Row(
        children: [
          Column(
            crossAxisAlignment: CrossAxisAlignment.start,
            children: [
              Text(_subHeading, style: Theme.of(context).textTheme.bodyLarge?.copyWith(fontWeight: FontWeight.w600, color: Colors.white)),
              Text(_heading.toUpperCase(), style: Theme.of(context).textTheme.headlineMedium?.copyWith(fontWeight: FontWeight.w400, color: Colors.white)),
            ],
          ),
          Expanded(child: Container()),
          _svg.isNullOrEmpty() ? SizedBox.shrink() : Center(child: SvgPicture.asset(_svg!, package: _svgPackage, height: 28, colorFilter: ColorFilter.mode(Colors.white, BlendMode.srcIn))),
        ],
      ),
      SizedBox(height: 16),
      Text(_message),
      Expanded(
        child: Align(
            alignment: Alignment.bottomCenter,
            child: Padding(
              padding: const EdgeInsets.only(bottom: 16.0),
              child: OutlinedButton(
                onPressed: _dataAvailable ? () {widget.dispatcher.close();} : null,
                style: ButtonStyle(
                    padding: WidgetStatePropertyAll(EdgeInsets.symmetric(horizontal: 48, vertical: 16)),
                    shape: WidgetStatePropertyAll(
                      RoundedRectangleBorder(
                        borderRadius: BorderRadius.circular(8),
                      ),
                    ),
                    foregroundColor: WidgetStateProperty.resolveWith(
                          (Set<WidgetState> states) {
                        return states.contains(WidgetState.disabled)
                            ? Colors.white.withValues(alpha: 0.8)
                            : Colors.white;
                      },
                    ),
                    side: WidgetStateProperty.resolveWith(
                          (Set<WidgetState> states) {
                        return states.contains(WidgetState.disabled)
                            ? BorderSide(color: Colors.white.withValues(alpha: 0.8), width: 2)
                            : BorderSide(color: Colors.white, width: 2);
                      },
                    )
                ),
                child: Text('CLOSE'),
              ),
            )
        ),
      )
    ];
  }

  List<Widget> _renderLoader() {
    return [
      Expanded(
        child: Center(
          child: SpacedColumn(
            spacing: 8,
            mainAxisAlignment: MainAxisAlignment.center,
            children: [
              CircularProgressIndicator(
                value: null,
                color: Colors.white,
              ),
              Text(_loadingStatusMessage)
            ],
          ),
        ),
      ),
      Expanded(child: Container())
    ];
  }

  @override
  Widget build(BuildContext context) {
    return DefaultTextStyle(
      style: TextStyle(inherit: true, color: Colors.white),
      child: ColoredBox(
        color: _getOverlayColor(),
        child: Stack(
          fit: StackFit.expand,
          children: [
            Padding(
              padding: EdgeInsets.only(top: MediaQuery.of(context).padding.top),
              child: Column(
                mainAxisAlignment: MainAxisAlignment.center,
                children: [
                  Expanded(
                    flex: 1,
                    child: Center(
                      child: Padding(
                        padding: const EdgeInsets.only(left: 32, right: 32, top: 32),
                        child: Column(
                          mainAxisAlignment: MainAxisAlignment.center,
                          children: _barcodeText.isNullOrEmpty() ? [] : [
                            SizedBox(
                              width: MediaQuery.of(context).size.width * 0.6,
                              height: 84,
                              child: BarcodeWidget(data: _barcodeText, barcode: BarcodeLib.Barcode.fromType(_barcodeType), color: Colors.white, drawText: false),
                            ),
                            SizedBox(height: 12),
                            Text(
                              _barcodeText.truncateWithEllipsis(120),
                              style: Theme.of(context).textTheme.titleMedium?.copyWith(color: Colors.white),
                              textAlign: TextAlign.center,
                            ),
                          ],
                        ),
                      ),
                    ),
                  ),
                  Expanded(
                    flex: 1,
                    child: Padding(
                      padding: EdgeInsets.fromLTRB(24, 24, 24, MediaQuery.of(context).padding.bottom),
                      child: Column(
                        mainAxisAlignment: MainAxisAlignment.start,
                        crossAxisAlignment: CrossAxisAlignment.start,
                        children: [
                          ...(_dataAvailable ? _renderOutcomeWidgets() : _renderLoader()),
                        ],
                      ),
                    ),
                  )
                ],
              ),
            )
          ],
        ),
      ),
    );
  }
}
